/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.client.console;

import de.bsvrz.dav.daf.main.ClientDavInterface;
import de.bsvrz.pua.prot.Protocols;
import de.bsvrz.pua.prot.manager.requests.DataManagerRequest;
import de.bsvrz.pua.prot.manager.taskmanager.TaskManagerInterface;
import de.bsvrz.pua.prot.manager.taskmanager.ThreadPool;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.CmdInterpreter;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.CmdMenu;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.Command;
import de.bsvrz.sys.funclib.losb.util.cmdinterface.InfoCommands;

import java.io.BufferedReader;
import java.io.BufferedWriter;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.net.Socket;
import java.net.UnknownHostException;

/**
 * Klasse der Pua-Steuerungskonsole. Erlaubt das Steuern von Pua über eine Telnet-Session.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
public class Console {
	/*
	 * WICHTIG
	 * Wenn zusätzliche Kommados / Menüs eingefügt werden, dann MUSS
	 * sendQuit(String, int) angepasst werden!
	 * Dort muss die Nummer des Menüpunktes 'Herunterfahren' eingetragen werden.  
	 */

	/** Timeout beim Beenden von Protokolle und Auswertungen. */
	private static int SEND_TIMEOUT = 60 * 5;

	/** Telnet-Konsole */
	private CmdInterpreter _cmd;

	/** Protocols */
	private Protocols _protocols;

	/**
	 * Erzeugt die Konsole.
	 *
	 * @param dav       Verbindung zum Datenverteiler.
	 * @param protocols Hauptklasse
	 * @param port      Port der Konsole.
	 * @param timeout   Timeout einer inaktiven Session in Minuten.
	 */
	public void createConsole(final ClientDavInterface dav, 
			final Protocols protocols, 
			final int port, 
			final int timeout) {
		this._protocols = protocols;
		_cmd = new CmdInterpreter(port, timeout);
		CmdMenu root = new CmdMenu("Protokolle und Auswertungen: Hauptmenue", "Hauptmenü");

		root.addNode(createQueueMenu());
		root.addNode(InfoCommands.getDebugLevelMenu());
		root.addNode(createInfoMenu());
		root.addCmd(new ShowSubscriptionsCommand(dav));
		root.addCmd(new ShutdownCommand(_protocols));

		_cmd.setMenu(root);
		_cmd.start();
	}


	/**
	 * Liefert Submenü für System-, Laufzeit- und Speicherinformationen.
	 *
	 * @return Menü
	 */
	private CmdMenu createInfoMenu() {
		CmdMenu im = new CmdMenu("Laufzeit- und Softwareinformationen", "");
		im.addCmd(InfoCommands.systemInfo());
		im.addCmd(InfoCommands.runtimeInfo());
		im.addCmd(InfoCommands.memoryInfo());
		return im;
	}

	/**
	 * Liefert ein Submenu zum Anzeigen der Queues zurück.
	 *
	 * @return Queues
	 */
	private CmdMenu createQueueMenu() {
		CmdMenu qm = new CmdMenu("Warteschlangen anzeigen", "");
		//Protokolle in Erstellung
		qm.addCmd(
				new Command("In Erstellung befindliche Protokolle anzeigen.", "") {
					@Override
					public void execute() throws Exception {
						this.println("Momentan werden erstellt / gesendet:");
						final TaskManagerInterface taskManager = _protocols._manager.getTaskManager();
						if(taskManager != null) {
							for(Long protocol : taskManager.getActiveProtocols()) {
								this.println("" + protocol);		//nicht .toString falls null
							}
						}
					}
				}
		);

		// Queue Taskmanager (Protokollwarteschlange)
		qm.addCmd(
				new Command("Protokollwarteschlange anzeigen.", "") {
					@Override
					public void execute() throws Exception {
						this.println("Demnächst werden erstellt:");
						final TaskManagerInterface taskManager = _protocols._manager.getTaskManager();
						if(taskManager != null) {
							for(final ThreadPool.QueuedThread thread : taskManager.getWaitingProtocols()) {
								this.println(thread.getRequestId() + " -> " + thread.getProtocolId());	
							}
						}
					}
				}
		);

		//Queue Datenmanager (Auftragswarteschlange)
		qm.addCmd(
				new Command("Auftragswarteschlange.", "") {
					@Override
					public void execute() throws Exception {
						this.println("Es stehen folgende Operationen an:");
						if(_protocols._manager.getDataManager() != null) {
							for(DataManagerRequest request : _protocols._manager.getDataManager().getUpcomingOps(new DataManagerRequest[0])) {
								this.println(request.toString());
							}
						}
					}
				}
		);

		return qm;
	}

	/**
	 * Beendet Protokolle und Auswertungen.
	 *
	 * @param host Host
	 * @param port Port
	 *
	 * @throws UnknownHostException Unbekannter Host
	 * @throws IOException          Ein-/Ausgabefehler
	 */
	public static void sendQuitCmd(String host, int port) throws UnknownHostException, IOException {
		Socket connection = new Socket(host, port);
		connection.setSoTimeout(SEND_TIMEOUT * 1000);
		BufferedWriter out = new BufferedWriter(new OutputStreamWriter(connection.getOutputStream()));
		BufferedReader in = new BufferedReader(new InputStreamReader(connection.getInputStream()));

		readPrompt(in);
		write(out, "5");	//<- hier die Nummer des Menüpunktes 'Herunterfahren' eintragen.
		readPrompt(in);
		write(out, "j");
		in.readLine();
	}

	private static void readPrompt(BufferedReader in) throws IOException {
		while(!in.readLine().endsWith(CmdInterpreter.PROMPT)) ;
	}

	private static void write(BufferedWriter out, String s) throws IOException {
		out.write(s);
		out.newLine();
		out.flush();
	}

	/**
	 * Beendet den Telnet-Server. Trennt dabei alle Verbindungen.
	 *
	 * @throws Exception Fehler.
	 */
	public void terminateServer() throws Exception {
		_cmd.stopServerCmd();
	}
}
