/*
 * Copyright 2005-2008 by beck et al. projects GmbH, Munich
 * Copyright 2009 by Kappich Systemberatung, Aachen
 * 
 * This file is part of de.bsvrz.pua.prot.
 * 
 * de.bsvrz.pua.prot is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * de.bsvrz.pua.prot is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with de.bsvrz.pua.prot.  If not, see <http://www.gnu.org/licenses/>.
 */

package de.bsvrz.pua.prot.util;

import de.bsvrz.dav.daf.main.config.Aspect;
import de.bsvrz.dav.daf.main.config.Attribute;
import de.bsvrz.dav.daf.main.config.AttributeGroup;
import de.bsvrz.dav.daf.main.config.ConfigurationException;
import de.bsvrz.dav.daf.main.config.DataModel;
import de.bsvrz.dav.daf.main.config.ObjectSet;
import de.bsvrz.dav.daf.main.config.ReferenceAttributeType;
import de.bsvrz.dav.daf.main.config.SystemObject;
import de.bsvrz.dav.daf.main.config.SystemObjectType;
import de.bsvrz.pua.prot.exceptions.CriticalParserException;
import de.bsvrz.pua.prot.exceptions.ScopeException;
import de.bsvrz.pua.prot.interpreter.InterpreterErrorMessage;
import de.bsvrz.pua.prot.interpreter.semantics.SymbolTable;
import de.bsvrz.sys.funclib.losb.messages.ErrorMessage;

import java.util.*;

/**
 * Stellt Methoden zur Verfügung, die überprüfen ob ein SystemObjekt einem anderen SystemObjekt lt. dem verwendeten Datenmodel zugeordnet werden kann.
 *
 * @author beck et al. projects GmbH
 * @author Martin Hilgers
 * @version $Revision$ / $Date$ / ($Author$)
 */
@SuppressWarnings("deprecation")
public class MemberCheck {

	/** Zeichen mit dem freie Aspekte beginnen */
	public static final char FREE_ASPECT_CHARACTER = '?';

	/** Objekt ist ein Typobjekt */
	public static final int OBJECTTYPE = 0;

	/** Objekt ist ein Objekt */
	public static final int OBJECT = 1;

	/** Objekt ist eine Attributgruppe */
	public static final int ATTRIBUTEGROUP = 2;

	/** Objekt ist ein Aspekt */
	public static final int ASPECT = 3;

	/** Objekt ist ein Attribut */
	public static final int ATTRIBUTE = 5;

	/** Objekt ist unbekannt */
	public static final int UNKNOWN = 6;

	/** Ungebundener Aspekt */
	public static final int FREE_ASPECT = 7;

	/** Interne Verwendung: Typ wurde noch nicht festgestellt */
	private static final int UNINITIALIZED = -1;

	/** Datenmodell */
	private DataModel _model;

	/** Symboltabelle */
	private SymbolTable _symbolTable;

	/**
	 * Konstruktor
	 *
	 * @param model Datenmodell
	 * @param st    Symboltabelle
	 */
	public MemberCheck(DataModel model, SymbolTable st) {
		this._model = model;
		_symbolTable = st;
	}

	/**
	 * Ausgabe eines Typs
	 *
	 * @param i Typ der als String zurückgegeben werden soll (ASPECT | ATTRIBUTE | ...
	 *
	 * @return String mit Typenbezeichnung (Aspect | Attribut | ...
	 */
	static public String toString(int i) {
		String r = "";

		switch(i) {
			case ASPECT:
				r = "Aspekt";
				break;
			case ATTRIBUTE:
				r = "Atttribut";
				break;
			case ATTRIBUTEGROUP:
				r = "Attributgruppe";
				break;
			case FREE_ASPECT:
				r = "ungebundener Aspekt";
				break;
			case OBJECT:
				r = "Objekt";
				break;
			case OBJECTTYPE:
				r = "Objekttyp";
				break;
		}

		return r;
	}

	/**
	 * Überprüft ob eine Attributgruppe durch ein Attribut referenziert wird
	 *
	 * @param attribute Attribut das auf PID referenzieren soll
	 * @param pid       PID des Objektes
	 *
	 * @return Objekt das durch PID bezeichnet wird, falls es von attribute referenziert wird
	 *
	 * @throws ScopeException          Falls der Objekttyp nicht durch das Attribut referenziert wird
	 * @throws CriticalParserException Falls es zu einem Kommuniktaionsfehler kommt
	 */
	public SystemObject isObjectReference(ReferenceAttributeType attribute, String pid) throws CriticalParserException {
		try {
			return isAttributeGroupMember(attribute.getReferencedObjectType(), _model.getAttributeGroup(pid));
		}
		catch(ConfigurationException ce) {
			throw new CriticalParserException(ErrorMessage.COMMUNICATION + ce);
		}
	}

	/**
	 * Überprüft ob ein Objekttyp von einem Attribut referenziert wird
	 *
	 * @param attribute Attribut das auf PID referenzieren soll
	 * @param pid       PID des Objektes
	 *
	 * @return Objekt das durch PID bezeichnet wird, falls es von attribute referenziert wird
	 *
	 * @throws ScopeException Falls der Objekttyp nicht von dem Attribut referenziert wird
	 */
	public SystemObjectType isObjectTypeReference(Attribute attribute, String pid) throws ScopeException {
		if(attribute == null || pid == null) return null;

		SystemObjectType result = null;

		try {
			//int type = whatIs(attribute.getDataModel(), pid);
			SystemObjectType obj = _model.getType(pid);

			List<ObjectSet> sets = attribute.getObjectSets();

			for(ObjectSet objectSet : sets) {
				List<SystemObject> objects = objectSet.getElements();
				for(SystemObject objElement : objects) {
					if(objElement.getId() == obj.getId()) {
						result = obj;
						break;
					}
				}
			}
		}
		catch(ConfigurationException e) {
			result = null;
		}

		if(result == null) throw new ScopeException(InterpreterErrorMessage.IS_NO_ELEMENT_OF + pid + ", " + attribute.getPid());

		return result;
	}

	/**
	 * Überprüft ob ein Attribut Element einer Attributgruppe ist
	 *
	 * @param atg  Attributgruppe
	 * @param name Name des Attributs
	 *
	 * @return Attribute-Objekt falls zur Attributgruppe passend. null sonst.
	 *
	 * @throws CriticalParserException Falls das Attribut nihct zu der Attributgruppe gehört
	 */
	static public Attribute isAttributeMember(AttributeGroup atg, String name) throws CriticalParserException {
		if(atg == null || name == null) throw new ScopeException(InterpreterErrorMessage.MISSING_ATTRIBUTEGROUP_OR_ATT);

		Attribute result = null;

		try {
			result = atg.getAttribute(name);
		}
		catch(ConfigurationException e) {
			throw new CriticalParserException(ErrorMessage.COMMUNICATION);
		}

		if(result == null) throw new ScopeException(InterpreterErrorMessage.IS_NO_ELEMENT_OF + name + ", " + atg.getPid());
		return result;
	}

	/**
	 * Überprüft ob ein Aspekt zu einer Attributgruppe gehört
	 *
	 * @param atg Attributgruppe
	 * @param asp Aspekt
	 *
	 * @return Attributgruppen-Objekt falls der Aspekt zur Attributgruppe passend ist. null sonst.
	 *
	 * @throws ScopeException Falls der Aspekt nicht für die Attributgruppe gültig ist
	 */
	public static AttributeGroup isAspectMember(AttributeGroup atg, Aspect asp) throws ScopeException {
		if(atg == null) {
			throw new ScopeException(InterpreterErrorMessage.ATTRIBUTEGROUP_MISSING);
		}
		else if(asp == null) throw new ScopeException(InterpreterErrorMessage.ASPECT_EXPECTED);

		AttributeGroup result = null;

		try {
			Collection<Aspect> aspList = atg.getAspects();

			for(Aspect aspElement : aspList) {
				if(asp.getId() == aspElement.getId()) {
					result = atg;
					break;
				}
			}
		}
		catch(ConfigurationException e) {
		}

		if(result == null) throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + asp.getPid() + ", " + atg.getPid());
		return result;
	}


	/**
	 * Überprüft ob eine Attributgruppe zu einem Objekt gehört
	 *
	 * @param obj Objekt
	 * @param atg Attributgruppe
	 *
	 * @return AttributeGroup Objekt falls Attributgruppe dem Objekt zugeordnet werden kann.
	 *
	 * @throws ScopeException Falls die Attributgruppe nicht zum Objekt gehört
	 */
	public static AttributeGroup isAttributeGroupMember(SystemObject obj, AttributeGroup atg) throws ScopeException {
		if(obj == null) {
			throw new ScopeException(InterpreterErrorMessage.OBJECT_OR_OBJECTTYPE_EXPECTED);
		}
		else if(atg == null) throw new ScopeException(InterpreterErrorMessage.ATTRIBUTEGROUP_MISSING);

		AttributeGroup result = null;

		try {
			List<AttributeGroup> atgList;
			if(obj instanceof SystemObjectType) {
				atgList = ((SystemObjectType)obj).getAttributeGroups();
			}
			else {
				atgList = obj.getType().getAttributeGroups();
			}

			if(atg != null) {
				for(AttributeGroup atgElement : atgList) {
					if(atg.getId() == atgElement.getId()) {
						result = atg;
						break;
					}
				}
			}
		}
		catch(Exception e) {
			throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + atg.getPid() + ", " + obj.getPid());
		}

		if(result == null) throw new ScopeException(InterpreterErrorMessage.INVALID_SCOPE + atg.getPid() + ", " + obj.getPid());
		return result;
	}


	/**
	 * Gibt den Typ des mit PID bezeichneten Objekts zurück. Achtung bei Rückgabewert ATTRIBUTE: Diese Methode prüft nicht nach, ob es dieses Attribut wirklich
	 * gibt. (das ist aus Performance-Gründen nicht möglich)
	 *
	 * @param pid PID des Objekts
	 *
	 * @return Typ des Objekts (OBJECTTYPE, OBJECT, ATTRIBUTEGROUP, ASPECT, FREE_ASPECT, SIMVAR, ATTRIBUTE, oder UNKNOWN)
	 *
	 * @throws CriticalParserException Fehler
	 */
	public int whatIs(String pid) throws CriticalParserException {
		int result = UNINITIALIZED;

		try {
			SystemObject obj = getObject(pid, false);
			if(obj != null) {
				result = whatIs(obj);
			}
			else {
				// if it couldn't be looked up in the model, it can only be an attribute or a free_aspect
				if(pid.charAt(0) == '?') {
					return FREE_ASPECT;
				}
				else {
					return ATTRIBUTE;
				}
			}
		}
		catch(Exception e) {
			result = UNINITIALIZED;
			throw new CriticalParserException(e.getMessage());
		}

		return result;
	}

	/**
	 * Gibt den Typ des Objekts zurück
	 *
	 * @param obj Objekt dessen Typ überprüft werden soll
	 *
	 * @return Typ des Objekts (OBJECTTYPE, OBJECT, ATTRIBUTEGROUP, ASPECT, SIMVAR ATTRIBUTE oder UNKNOWN)
	 */
	static public int whatIs(SystemObject obj) {
		int result = UNKNOWN;

		if(obj instanceof Aspect) {
			result = ASPECT;
		}
		else if(obj instanceof Attribute) {
			result = ATTRIBUTE;
		}
		else if(obj instanceof AttributeGroup) {
			result = ATTRIBUTEGROUP;
		}
		else if(obj instanceof SystemObjectType) {
			result = OBJECTTYPE;
		}
		else {
			result = OBJECT;
		}

		return result;
	}


	/**
	 * Gibt das der übergebenen PID zugeordnete Objekt zurück. Falls es sich um einen Objekttyp handelt, wird das jeweilige Objekt zurückgegeben
	 *
	 * @param pid PID des gewünschten Objekts oder des Objekttyps
	 *
	 * @return Das jeweilige Objekt oder den Objekttyp, falls es keine Bindung gibt, oder null bei einer ungültigen PID
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt
	 */
	public SystemObject getObject(String pid) throws CriticalParserException {
		return getObject(pid, true);
	}

	/**
	 * Ersetzt einen Objekttyp durch ein Objekt. Falls es sich um einen Objekttyp handelt, wird das zugeordnete Objekt zurückgegeben. Falls es sich um ein Objekt
	 * handelt, wird dieses Objekt wieder zurückgegeben.
	 *
	 * @param obj Objekt oder Objekttyp.
	 *
	 * @return Das jeweilige Objekt, den Objekttyp, falls es keine Bindung gibt, oder null bei einer ungültigen PID
	 */
	public SystemObject getObject(SystemObject obj) {
		SystemObject result = obj;
		if(obj instanceof SystemObjectType) {
			SystemObjectType sot = (SystemObjectType)obj;
			result = _symbolTable.getObject(sot);
		}
		return result;
	}

	/**
	 * Gibt das der übergebenen PID zugeordnete Objekt zurück.
	 *
	 * @param pid            PID des gewünschten Objekts oder des Objekttyps
	 * @param useSymbolTable True: Verwende die Symboltabelle um das an den Objekttyp zu bindende Objekt zurückzugeben, falls pid einen Objekttyp bezeichnet
	 *
	 * @return Das jeweilige Objekt bzw. der Objekttyp oder null bei einer ungültigen PID
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt
	 */
	private SystemObject getObject(String pid, boolean useSymbolTable) throws CriticalParserException {
		SystemObject result = null;
		try {
			result = _model.getObject(pid);
		}
		catch(ConfigurationException e) {
			throw new CriticalParserException(ErrorMessage.COMMUNICATION);
		}
		if(result != null && result instanceof SystemObjectType && useSymbolTable) {
			result = _symbolTable.getObject((SystemObjectType)result);
		}
		return result;
	}

	/**
	 * Gibt den Objekttyp mit der übergebenen PID zurück
	 *
	 * @param pid PID des gewünschten Objekttyps
	 *
	 * @return Der Objekttyp oder null bei einer ungültigen PID
	 *
	 * @throws CriticalParserException Falls es zu einem Kommunikationsfehler kommt
	 */
	public SystemObjectType getType(String pid) throws CriticalParserException {
		SystemObject so = getObject(pid, false);
		if(so instanceof SystemObjectType) {
			return (SystemObjectType)so;
		}
		else {
			return null;
		}
	}


	/**
	 * Gibt den der übergebenen PID zugeordneten Aspekt zurück
	 *
	 * @param pid PID des gewünschten Objekts
	 *
	 * @return Das jeweilige Objekt oder null bei einer ungültigen PID (z.B. bei einem freien Aspekt)
	 *
	 * @throws CriticalParserException Fehler.
	 */
	public Aspect getAspect(String pid) throws CriticalParserException {
		Aspect result = null;
		try {
			result = (Aspect)getObject(pid);
		}
		catch(ClassCastException c) {
			throw new ScopeException(InterpreterErrorMessage.ASPECT_EXPECTED + pid);
		}
		return result;
	}

	/**
	 * Gibt die der übergebenen PID zugeordnete Attributgruppe zurück
	 *
	 * @param pid PID des gewünschten Objekts
	 *
	 * @return Das jeweilige Objekt oder null bei einer ungültigen PID
	 *
	 * @throws ScopeException          Bei einem Fehler
	 * @throws CriticalParserException Falls es zu einem Kommuniktaionsfehler kommt
	 */
	public AttributeGroup getAttributeGroup(String pid) throws CriticalParserException, ScopeException {
		AttributeGroup result = null;
		try {
			result = (AttributeGroup)getObject(pid);
			if(result == null) throw new ScopeException(InterpreterErrorMessage.ATTRIBUTEGROUP_EXPECTED + pid);
		}
		catch(ClassCastException c) {
			throw new ScopeException(InterpreterErrorMessage.ATTRIBUTEGROUP_EXPECTED + pid);
		}
		return result;
	}
}
